"""Common util functions used by other packages in tealer.

Modules:
    code_complexity: Contains util functions to compute code complexity
        of the contracts.

    command_line: Defines output util functions used by command line
        argument handlers.

    comparable_enum: Defines a comparable enum class.

    output: Defines util functions to output dot files and util
        classes to store, represent detector results.
"""
