"""Defines exceptions classes for representing exceptions in tealer.

* TealerException: Exception class for common exceptions in tealer.
"""


class TealerException(Exception):
    """Exception class to represent common exceptions in tealer"""
