"""Contains functions, classes for parsing and representing teal contracts.

Module contains implementations for representing and parsing teal contracts.

Additional to instructions, this modules contains implementation of Teal, BasicBlock
classes. BasicBlock class is used to represent a basic block in the teal contract.
Teal class is used to represent the teal contract. Teal class also comes with
api methods to run detectors, printers on the contract.

Modules:
    instructions: Contains functions, classes for parsing and representing teal
        instructions.

    basic_blocks: Defines class to represent basic blocks of teal contracts.

    global_field: Defines classes to represent global fields.

    parse_teal: Defines parsing functions for teal contracts.

    teal: Defines class to represent Teal contracts.
"""
