"""Static analyzer for Algorand Smart Contracts.

Tealer comes with the utilities to parse and represent
smart contracts written in TEAL. Tealer also contains
detectors that analyze the TEAL contracts and printers
that print interesting information extracted from the
contract.

Modules:
    detectors: Module for tealer detectors.

    printers: Module for tealer printers

    teal: Contains functions, classes for parsing and representing teal contracts.

    utils: Common util functions used by other packages in Tealer.

    exceptions: Defines exceptions classes for representing exceptions in tealer.
"""
