"""Contains functions, classes for parsing and representing teal instructions.

For representing teal code, this module contains classes for teal
instructions, transaction fields, App parameter fields, Asset parameter
fields and Asset holding fields.

This module also contains implementations of parsers for instructions,
global fields, transaction fields, App parameter fields,
Asset parameter fields and Asset holding fields.

Modules:
    app_params_field: Defines classes to represent app parameter fields.

    asset_holding_field: Defines classes to represent asset holding fields.

    asset_params_field: Defines classes to represent asset parameter fields.

    instructions: Defines classes to represent teal instructions.

    parse_app_params_field: Parser for app parameter fields.

    parse_asset_holding_field: Parser for asset holding fields.

    parse_asset_params_field: Parser for asset parameter fields.

    parse_global_field: Parser for global fields.

    parse_instruction: Parser for teal instructions.

    parse_transaction_field: Parser for transaction fields.

    transaction_field: Defines classes to represent transaction fields.
"""
